## Version 1.0 -- 2013 Jul 10

# Add the checks on the parameters

######################################################################
## SECTION (SPECIFICATION) ===========================================
######################################################################

## =======================
## SPECIFICATION -- INPUTS

## (nbVar) Number of variables x0, x1, ..., xn
##

## (nbReal) Number of samples to draw from the joint probability
## distribution of the input quantities

## (expression) Formula written in R that expresses an output quantity y
## as a function of nbVar input quantities x0, x1, ..., xn

## (type) List of integer corresponding to the Names of the families of
## probability distributions to be assigned to the input quantities
## selected from drop-down menus that include the families
## of distributions specified under SPECIFICATION -- DISTRIBUTIONS


## (parameters) Values of parameters sufficient to determine particular
## elements of the families of distributions selected in (I2) --
## refer to SPECIFICATION -- PARAMETERS for details on how this
## is to be done


## (correlation) The correlation coefficients
## between all pairs of input quantities

## (I6) If box (I4) is checked, then specification of a copula that
## defines a joint probability distribution for the input
## variables, consistent with the selections made in (I2), (I3),
## and (I5)
##

## (outputFile) The name of an output file where to save the data
truncnormL = require(truncnorm)
triangleL = require(triangle)
nleqslvL = require(nleqslv)
mvtnormL = require(mvtnorm)
numDerivL = require(numDeriv)


signifDigits = function (x, y)
{
	n = 1
	if(x==y)
		return(Inf)
	while (signif(x, n) == signif(y, n)) {n = n + 1}
	return(n-1)
}

gummer = function (f, x, u, r=diag(length(u)), ...)
{
	
	g = grad(f, x, ...)
	y = f(x, ...)
	uy = sqrt(matrix(g, nrow=1) %*% (outer(u,u,"*")*r) %*% matrix(g, ncol=1))
	return(c(y, uy))
}

distrib = function(nbVar,nbReal,varNames,expression,type,parameters,correlation,outputFile,copulaType=NULL,copulaDf=NULL)
{
	
	##	cat("nbVar = ",nbVar,"\n")
	##	cat("nbReal = ",nbReal,"\n")
	##	cat("expre = ",expression,"\n")
	##	cat("type = ",type,"\n")
	##	cat("parameters = \n")
	##	print(parameters)
	##	cat("correlation = \n")
	##	print(correlation)
	##	cat("outputFile = ",outputFile,"\n")
	
	## Verification that all the packages were loaded successfully
	
	if (!truncnormL)
		stop("Impossible to load the package truncnorm")
	if (!triangleL)
		stop("Impossible to load the package triangle")
	if (!nleqslvL)
		stop("Impossible to load the package nleqslv")
	if (!mvtnormL)
		stop("Impossible to load the package mvtnorm")
	if (!numDerivL)
		stop("Impossible to load the package numDeriv")


	## (COMP-03) Verify that the user has specified a positive
	## integer (m) for the size of the Monte Carlo sample.
	
	
	if(nbReal<=0)
		stop("Negative number or realisation",call.=FALSE)
	
	
	## (COMP-04) If correlations are given, verify that the resulting
	## correlation matrix has all eigenvalues positive.
	if(correlation)
	{
#		cat("copulaType = ",copulaType,"\n")
		if (min(eigen(correlation, symmetric=TRUE, only.values=TRUE)$values) < sqrt(.Machine$double.eps))
			stop("Illegal correlation matrix\n",call. = FALSE)
#		if(copulaType=="Student")
#			cat("copulaDf = ",copulaDf,"\n")
	}
	
	##Parameters are valid so the computation continues
	varEnv = new.env()
	meanY = numeric(2)
	medianY = numeric(2)
	sdY = numeric(2)
	madY = numeric(2)
	quantileY = rbind(quantile(0, probs=c(0.005, 0.025, 0.05,
							pnorm(-1), 0.25, 0.75, pnorm(1), 0.95, 0.975, 0.995)),0)
	
	sdX = numeric(nbVar)
	meanX = numeric(nbVar)
	
	for (irun in 1:2)
	{
		if(!correlation)
		{
			for(ii in 1:nbVar)
			{
				if(type[ii]==0)  ## Bernoulli (Prob. of success)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][1]>1 )
						stop(paste("In quantity ",varNames[ii]," Prob. of success must be between 0 and 1"),call.=FALSE)
					varTemp=rbinom(nbReal, size=1, prob=parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii]=parameters[[ii]][1];
						sdX[ii]=sqrt(parameters[[ii]][1]*(1-parameters[[ii]][1]));
					}
				}
				if(type[ii]==1)  ## Beta (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean*(Mean*(1-Mean)-StdDev^2)/StdDev^2
					b = (1-Mean)*(Mean*(1-Mean)-StdDev^2)/StdDev^2
					if(a<0 || b<0)
						stop(paste("In quantity ",varNames[ii]," Those paremeters are not suitable for a beta distribution"),call.=FALSE)
					varTemp=rbeta(nbReal, shape1=a, shape2=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
					
				}
				if(type[ii]==2) ## Beta (Shape1, Shape2)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The two shape parameters must be positive"),call.=FALSE)
					varTemp=rbeta(nbReal, shape1=parameters[[ii]][1], shape2=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1]/(parameters[[ii]][1]+parameters[[ii]][2])
						sdX[ii] = sqrt(parameters[[ii]][1]*parameters[[ii]][2]/(((parameters[[ii]][1]+parameters[[ii]][2])^2)*(parameters[[ii]][1]+parameters[[ii]][2]+1)))
					}
				}
				if(type[ii]==3) ## Chi-Squared (DF)
				{
					if(parameters[[ii]][1]<0)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be positive"),call.=FALSE)
					varTemp=rchisq(nbReal, df=parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = sqrt(2*parameters[[ii]][1]);
					}
				}
				if(type[ii]==4) ## Exponential (Mean)
				{
					varTemp= rexp(nbReal, rate=1/parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][1];
					}
				}
				if(type[ii]==5) ## Gamma (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][1]<=0)
						stop(paste("In quantity ",varNames[ii]," The mean must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					Shape = (Mean/StdDev)^2
					Scale = StdDev^2/Mean
					varTemp=rgamma(nbReal, shape=Shape, scale=Scale)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==6) ## Gamma (Shape, Scale)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The shape and the scale must be positive"),call.=FALSE)
					varTemp=rgamma(nbReal, shape=parameters[[ii]][1], scale=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1]* parameters[[ii]][2];
						sdX[ii]   = sqrt(parameters[[ii]][1])* parameters[[ii]][2];
					}
				}
				if(type[ii]==7) ## Gaussian (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					varTemp=rnorm(nbReal, mean=parameters[[ii]][1], sd=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==8) ## Gaussian -- Truncated (Mean, StdDev, Left, Right)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][3]>parameters[[ii]][4])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=rtruncnorm(nbReal, a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = etruncnorm(a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
						sdX[ii]   = sqrt(vtruncnorm(a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2]))
					}
				}
				if(type[ii]==9) ## Rectangular -- Continuous (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean - StdDev*sqrt(3)
					b = Mean + StdDev*sqrt(3)
					varTemp=runif(nbReal, min=a, max=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==10) ## Rectangular -- Continuous (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=runif(nbReal, min=parameters[[ii]][1], max=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = (parameters[[ii]][1]+parameters[[ii]][2])/2;
						sdX[ii]   = (parameters[[ii]][2]-parameters[[ii]][2])/sqrt(12);
					}
				}
				if(type[ii]==11) ## Student t (Mean, StdDev, DF)
				{
					if(parameters[[ii]][3]<2)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be greater than 2"),call.=FALSE)
					scale = parameters[[ii]][2]/sqrt(parameters[[ii]][3]/(parameters[[ii]][3]-2))
					varTemp = parameters[[ii]][1] + scale*rt(nbReal, df=parameters[[ii]][3])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii] = parameters[[ii]][2];
					}
				}
				if(type[ii]==12) ## Student t (Center, Scale, DF) 
				{
					if(parameters[[ii]][3]<0)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be positive"),call.=FALSE)
					varTemp = parameters[[ii]][1] + parameters[[ii]][2]*rt(nbReal, df=parameters[[ii]][3])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						if (parameters[[ii]][3] > 2)
							sdX[ii] = parameters[[ii]][2] * sqrt(parameters[[ii]][3]/(parameters[[ii]][3]-2))
						else
							sdX[ii] = Inf
						
					}
				}
				
				
				if(type[ii]==13) ## Triangular -- Symmetric (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					c = Mean
					a = Mean - StdDev*sqrt(6)
					b = Mean + StdDev*sqrt(6)
					varTemp = rtriangle(nbReal, a, b, c)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				
				if(type[ii]==14) ## Triangular -- Symmetric (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp = rtriangle(nbReal, a=parameters[[ii]][1], b=parameters[[ii]][2], c=(parameters[[ii]][1]+parameters[[ii]][2])/2)
					if(irun==1)
					{
						a = parameters[[ii]][1]
						b = parameters[[ii]][2]
						c = (a+b)/2
						meanX[ii] = (a+b+c)/3;
						sdX[ii]   = sqrt((a^2 + b^2 + c^2 - a*b - a*c - b*c)/18)
					}
				}
				if(type[ii]==15) ## Triangular -- Asymmetric (Left, Mode, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2] || parameters[[ii]][2]>parameters[[ii]][3])
						stop(paste("In quantity ",varNames[ii]," Parameters must respect Left<=Mode<=Right "),call.=FALSE)
					varTemp = rtriangle(nbReal, a=parameters[[ii]][1], b=parameters[[ii]][3], c=parameters[[ii]][2])
					if(irun==1)
					{
						a = parameters[[ii]][1]
						b = parameters[[ii]][3]
						c = parameters[[ii]][2]
						meanX[ii] = (a+b+c)/3;
						sdX[ii]   = sqrt((a^2 + b^2 + c^2 - a*b - a*c - b*c)/18)
					}
				}
				if(type[ii]==16) ## Uniform -- Continuous (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean - StdDev*sqrt(3)
					b = Mean + StdDev*sqrt(3)
					varTemp=runif(nbReal, min=a, max=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==17) ## Uniform -- Continuous (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=runif(nbReal, min=parameters[[ii]][1], max=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = (parameters[[ii]][1]+parameters[[ii]][2])/2;
						sdX[ii]   = (parameters[[ii]][2]-parameters[[ii]][2])/sqrt(12);
					}
				}
				if(type[ii]==18) ## Weibull (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][1]<=0)
						stop(paste("In quantity ",varNames[ii]," The mean must be positive"),call.=FALSE)
					f = function (betaeta, mu, sigma)
					{
						beta = betaeta[1]; eta = betaeta[2]
						y = numeric(2)
						y[1] = eta*gamma(1 + 1/beta) - mu
						y[2] = eta^2 * (gamma(1 + 2/beta) - gamma(1 + 1/beta)^2) - sigma^2
						return(y)
					}
					s = nleqslv(x=c(1,1), fn=f, mu=parameters[[ii]][1], sigma=parameters[[ii]][2])
					if (s$termcd %in% c(1,2)) {beta = s$x[1]; eta = s$x[2]} else {
						cat("Cannot assign given Mean and StdDev to Weibull distribution\n")}
					varTemp=rweibull(nbReal, shape=beta, scale=eta)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==19) ## Weibull (Shape, Scale)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The shape and the scale must be positive"),call.=FALSE)
					varTemp=rweibull(nbReal, shape=parameters[[ii]][1], scale=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][2] * gamma(1 + 1/parameters[[ii]][1]);
						sdX[ii]   = parameters[[ii]][2] * sqrt(gamma(1 + 2/parameters[[ii]][1]) - (gamma(1 + 1/parameters[[ii]][1]))^2);
					}
				}
				if(type[ii]==20) ## von Mises (Mean, Concentration)
				{
					varTemp=rnorm(nbReal, mean=parameters[[ii]][1], sd=parameters[[ii]][2])
#					varTemp=rvm(nbReal, parameters[[ii]][1], k=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = 1/parameters[[ii]][2];
					}
				}
				
				
				eval(parse(text=paste(varNames[ii], "=varTemp",sep="")),envir = varEnv)
				
			}
			
		}else
		
		{
			
			if(copulaType=="Gaussian")
			{
				z = rmvnorm(nbReal, mean=rep(0, nbVar), sigma=correlation)
				z = pnorm(z)
			}else
			{
				z = rmvt(nbReal, sigma=correlation, df=copulaDf)
				z = pt(z, df=copulaDf)
			}
			
			
			for(ii in 1:nbVar)
			{
				if(type[ii]==0)  ## Bernoulli (Prob. of success)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][1]>1 )
						stop(paste("In quantity ",varNames[ii]," Prob. of success must be between 0 and 1"),call.=FALSE)
					varTemp=qbinom(z[,ii], size=1, prob=parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii]=parameters[[ii]][1];
						sdX[ii]=sqrt(parameters[[ii]][1]*(1-parameters[[ii]][1]));
					}
				}
				if(type[ii]==1)  ## Beta (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean*(Mean*(1-Mean)-StdDev^2)/StdDev^2
					b = (1-Mean)*(Mean*(1-Mean)-StdDev^2)/StdDev^2
					if(a<0 || b<0)
						stop(paste("In quantity ",varNames[ii]," Those paremeters are not suitable for a beta distribution"),call.=FALSE)
					varTemp=qbeta(z[,ii], shape1=a, shape2=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
					
				}
				if(type[ii]==2) ## Beta (Shape1, Shape2)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The two shape parameters must be positive"),call.=FALSE)
					varTemp=qbeta(z[,ii], shape1=parameters[[ii]][1], shape2=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1]/(parameters[[ii]][1]+parameters[[ii]][2])
						sdX[ii] = sqrt(parameters[[ii]][1]*parameters[[ii]][2]/(((parameters[[ii]][1]+parameters[[ii]][2])^2)*(parameters[[ii]][1]+parameters[[ii]][2]+1)))
					}
				}
				if(type[ii]==3) ## Chi-Squared (DF)
				{
					if(parameters[[ii]][1]<0)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be positive"),call.=FALSE)
					varTemp=qchisq(z[,ii], df=parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = sqrt(2*parameters[[ii]][1]);
					}
				}
				if(type[ii]==4) ## Exponential (Mean)
				{
					varTemp= qexp(z[,ii], rate=1/parameters[[ii]][1])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][1];
					}
				}
				if(type[ii]==5) ## Gamma (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][1]<=0)
						stop(paste("In quantity ",varNames[ii]," The mean must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					Shape = (Mean/StdDev)^2
					Scale = StdDev^2/Mean
					varTemp=qgamma(z[,ii], shape=Shape, scale=Scale)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==6) ## Gamma (Shape, Scale)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The shape and the scale must be positive"),call.=FALSE)
					varTemp=qgamma(z[,ii], shape=parameters[[ii]][1], scale=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1]* parameters[[ii]][2];
						sdX[ii]   = sqrt(parameters[[ii]][1])* parameters[[ii]][2];
					}
				}
				if(type[ii]==7) ## Gaussian (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					varTemp=qnorm(z[,ii], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==8) ## Gaussian -- Truncated (Mean, StdDev, Left, Right)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][3]>parameters[[ii]][4])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=qtruncnorm(z[,ii], a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = etruncnorm(a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
						sdX[ii]   = sqrt(vtruncnorm(a=parameters[[ii]][3], b=parameters[[ii]][4], mean=parameters[[ii]][1], sd=parameters[[ii]][2]))
					}
				}
				if(type[ii]==9) ## Rectangular -- Continuous (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean - StdDev*sqrt(3)
					b = Mean + StdDev*sqrt(3)
					varTemp=qunif(z[,ii], min=a, max=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==10) ## Rectangular -- Continuous (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=qunif(z[,ii], min=parameters[[ii]][1], max=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = (parameters[[ii]][1]+parameters[[ii]][2])/2;
						sdX[ii]   = (parameters[[ii]][2]-parameters[[ii]][2])/sqrt(12);
					}
				}
				if(type[ii]==11) ## Student t (Mean, StdDev, DF)
				{
					if(parameters[[ii]][3]<2)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be greater than 2"),call.=FALSE)
					scale = parameters[[ii]][2]/sqrt(parameters[[ii]][3]/(parameters[[ii]][3]-2))
					varTemp = parameters[[ii]][1] + scale*qt(z[,ii], df=parameters[[ii]][3])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii] = parameters[[ii]][2];
						
					}
				}
				if(type[ii]==12) ## Student t (Center, Scale, DF)
				{
					if(parameters[[ii]][3]<0)
						stop(paste("In quantity ",varNames[ii]," No. of degrees of freedom must be positive"),call.=FALSE)
					varTemp = parameters[[ii]][1] + parameters[[ii]][2]*qt(z[,ii], df=parameters[[ii]][3])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						if (parameters[[ii]][3] > 2)
							sdX[ii] = parameters[[ii]][2] * sqrt(parameters[[ii]][3]/(parameters[[ii]][3]-2))
						else
							sdX[ii] = Inf
						
					}
				}
				if(type[ii]==13) ## Triangular -- Symmetric (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					c = Mean
					a = Mean - StdDev*sqrt(6)
					b = Mean + StdDev*sqrt(6)
					varTemp = qtriangle(z[,ii], a, b, c)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==14) ## Triangular -- Symmetric (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp = qtriangle(z[,ii], a=parameters[[ii]][1], b=parameters[[ii]][2], c=(parameters[[ii]][1]+parameters[[ii]][2])/2)
					if(irun==1)
					{
						a = parameters[[ii]][1]
						b = parameters[[ii]][2]
						c = (a+b)/2
						meanX[ii] = (a+b+c)/3;
						sdX[ii]   = sqrt((a^2 + b^2 + c^2 - a*b - a*c - b*c)/18)
					}
				}
				if(type[ii]==15) ## Triangular -- Asymmetric (Left, Mode, Right) 
				{if(parameters[[ii]][1]>parameters[[ii]][2] || parameters[[ii]][2]>parameters[[ii]][3])
						stop(paste("In quantity ",varNames[ii]," Parameters must respect Left<=Mode<=Right "),call.=FALSE)
					varTemp = qtriangle(z[,ii], a=parameters[[ii]][1], b=parameters[[ii]][3], c=parameters[[ii]][2])
					if(irun==1)
					{
						a = parameters[[ii]][1]
						b = parameters[[ii]][3]
						c = parameters[[ii]][2]
						meanX[ii] = (a+b+c)/3;
						sdX[ii]   = sqrt((a^2 + b^2 + c^2 - a*b - a*c - b*c)/18)
					}
				}
				if(type[ii]==16) ## Uniform -- Continuous (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					Mean=parameters[[ii]][1]
					StdDev=parameters[[ii]][2]
					a = Mean - StdDev*sqrt(3)
					b = Mean + StdDev*sqrt(3)
					varTemp=qunif(z[,ii], min=a, max=b)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==17) ## Uniform -- Continuous (Left, Right)
				{
					if(parameters[[ii]][1]>parameters[[ii]][2])
						stop(paste("In quantity ",varNames[ii]," Left must be less than or equal to Right "),call.=FALSE)
					varTemp=qunif(z[,ii], min=parameters[[ii]][1], max=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = (parameters[[ii]][1]+parameters[[ii]][2])/2;
						sdX[ii]   = (parameters[[ii]][2]-parameters[[ii]][2])/sqrt(12);
					}
				}
				if(type[ii]==18) ## Weibull (Mean, StdDev)
				{
					if(parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," StdDev must be positive"),call.=FALSE)
					if(parameters[[ii]][1]<=0)
						stop(paste("In quantity ",varNames[ii]," The mean must be positive"),call.=FALSE)
					f = function (betaeta, mu, sigma)
					{
						beta = betaeta[1]; eta = betaeta[2]
						y = numeric(2)
						y[1] = eta*gamma(1 + 1/beta) - mu
						y[2] = eta^2 * (gamma(1 + 2/beta) - gamma(1 + 1/beta)^2) - sigma^2
						return(y)
					}
					s = nleqslv(x=c(1,1), fn=f, mu=parameters[[ii]][1], sigma=parameters[[ii]][2])
					if (s$termcd %in% c(1,2)) {beta = s$x[1]; eta = s$x[2]} else {
						cat("Cannot assign given Mean and StdDev to Weibull distribution\n")}
					varTemp=qweibull(z[,ii], shape=beta, scale=eta)
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = parameters[[ii]][2];
					}
				}
				if(type[ii]==19) ## Weibull (Shape, Scale)
				{
					if(parameters[[ii]][1]<0 || parameters[[ii]][2]<0)
						stop(paste("In quantity ",varNames[ii]," The shape and the scale must be positive"),call.=FALSE)
					varTemp=qweibull(z[,ii], shape=parameters[[ii]][1], scale=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][2] * gamma(1 + 1/parameters[[ii]][1]);
						sdX[ii]   = parameters[[ii]][2] * sqrt(gamma(1 + 2/parameters[[ii]][1]) - (gamma(1 + 1/parameters[[ii]][1]))^2);
					}
				}
				if(type[ii]==20) ## von Mises (Mean, Concentration)
				{
					varTemp=qnorm(z[,ii], mean=parameters[[ii]][1], sd=parameters[[ii]][2])
#					varTemp=rvm(nbReal, parameters[[ii]][1], k=parameters[[ii]][2])
					if(irun==1)
					{
						meanX[ii] = parameters[[ii]][1];
						sdX[ii]   = 1/parameters[[ii]][2];
					}
				}
				
				eval(parse(text=paste(varNames[ii], "=varTemp",sep="")),envir = varEnv)
				if(irun==1)
				{
					meanX[ii]=mean(varTemp);
					sdX[ii]=sd(varTemp);
#					cat("\nVariable ",varNames[ii] ,"= \n")
#					print(head(varTemp))
#					cat("Mean =  ",mean(varTemp), " StdDev = ",sd(varTemp) ,"\n")
				}
			}
			
		}
		y=NULL
#		if(exists("pretreatment"))
#			print(pretreatment)
#		print(paste( "y =",expression,sep=""))
		if(exists("pretreatment"))
			try(eval(parse(text=pretreatment),envir = varEnv),silent = TRUE)
		try(eval(parse(text=paste( "y =",expression,sep="")),envir = varEnv),silent = TRUE)
		y = eval(parse(text="y"),envir = varEnv)
		
		if(is.null(y))
		{
			if(exists("pretreatment"))
				try(eval(parse(text=pretreatment),envir = varEnv))
			try(eval(parse(text=paste( "y =",expression,sep="")),envir = varEnv))
			stop(paste("Error in the evaluation of the expression:" ,expression, "\n"),call.=FALSE)
		}
		
		if(irun==1)
		{
			x11()
			plot(density(y, from=mean(y)-3*sd(y), to=mean(y)+3*sd(y)), bty="n",
					col="Blue", main="",
					xlab="Output quantity (Y)", ylab="Probability Density")
			curve(dnorm(x, mean=mean(y), sd=sd(y)), lty=3, lwd=2, col="Red", from=mean(y)-3*sd(y), to=mean(y)+3*sd(y), add=TRUE)
#			cat("\ny = \n")
#			print(head(y))
			if(outputFile!="no-output")
			{
				tempout=unlist(strsplit(outputFile,split="\\."))
				if (tempout[length(tempout)]!="txt")
					outputFile=paste(outputFile,"txt",sep=".")
				tempout=unlist(strsplit(outputFile,split="\\."))
				tempout2=tempout[1]
				if(length(tempout)>2)
					for(iu in 2:(length(tempout)-1))
						tempout2=paste(tempout2,tempout[iu],sep=".")
				outputFile =paste(tempout2,"-values.txt",sep="")
				outputFile2=paste(tempout2,"-results.txt",sep="")
				outputFile3=paste(tempout2,"-density.jpg",sep="")
				jpeg(file=outputFile3, width=5, height=4, units="in",
						res=600, quality=100, pointsize=10, bg="white")
				par(mar=c(4.5, 4.5, 0.5, 0.75))
				plot(density(y, from=mean(y)-3*sd(y), to=mean(y)+3*sd(y)), bty="n",
						col="Blue", main="",
						xlab="Output quantity (Y)", ylab="Probability Density")
				curve(dnorm(x, mean=mean(y), sd=sd(y)), lty=3, lwd=2, col="Red", from=mean(y)-3*sd(y), to=mean(y)+3*sd(y), add=TRUE)
				dev.off()
				cat("\nY values saved in file:  ",outputFile,"\n")
				cat("Results saved in file:   ",outputFile2,"\n")
				cat("Y Density saved in file: ",outputFile3,"\n")
				
			}
		}
		
		meanY[irun]=mean(y)
		medianY[irun]=median(y)
		sdY[irun]=sd(y)
		madY[irun]=mad(y)
		quantileY[irun,] =
				quantile(y, probs=c(0.005, 0.025, 0.05,
								pnorm(-1), 0.25, 0.75, pnorm(1), 0.95,
								0.975, 0.995))
		if(irun==1 && outputFile!="no-output")
		{
			write(t(matrix(y,ncol=1)), file = outputFile,	append = FALSE, sep = "\n")
		}
		
	}
	
	
	functionString = "yfunction = function (x) {"
	for(ik in 1:nbVar)
		functionString = paste(functionString, varNames[ik] ,"=x[",ik,"];")
	if(exists("pretreatment"))
		functionString = paste(functionString, pretreatment)
	functionString = paste(functionString, "return(",expression,")}")
#	print(functionString)
	try(eval(parse(text=functionString)))
	
	if(exists("yfunction"))
	{
		if(!correlation)
			sdGauss = gummer(f=yfunction, x=meanX, u=sdX)
		else
			sdGauss = gummer(f=yfunction, x=meanX, u=sdX, r=correlation)
	}
	
	## Sensitivity Coefficients and Uncertainty Budget
	ygrad = grad(f=yfunction, x=meanX)
	sensitivity = ygrad
	if(is.na(sdGauss[2]) || (sdGauss[2]==Inf || sdGauss[2]==0 ))
	{
		proportions = rep(NA,length(sdX)+1)
	} else
	{
		residual = 1-sum(((ygrad*sdX)^2)/sdGauss[2]^2)
		if (abs(residual) < sqrt(.Machine$double.eps)) {residual = 0}
		proportions = c(((ygrad*sdX)^2)/(sdGauss[2]^2), residual)
	}
	
	budget = data.frame(
			"SensitivityCoeffs"=signif(c(sensitivity, NA), 2),
			"Percent.u2"=signif(100*proportions, 2))
	rownames(budget) = c(varNames, "Correlations")
	
	cat("\n\n===== RESULTS ==============================\n\n")
	cat(    "Monte Carlo Method\n\n")
	cat("Summary statistics for sample of size", nbReal, "\n\n")
	cat("ave     =", signif(meanY[1],
					signifDigits(meanY[1], meanY[2])), "\n")
	cat("sd      =", signif(sdY[1], signifDigits(sdY[1], sdY[2])), "\n")
	cat("median  =",
			signif(medianY[1], signifDigits(medianY[1], medianY[2])), "\n")
	cat("mad     =", signif(madY[1],
					signifDigits(madY[1], madY[2])), "\n\n")
	
	cat("Coverage intervals\n\n")
	L = signif(quantileY[1,1],
			signifDigits(quantileY[1,1], quantileY[2,1]))
	R = signif(quantileY[1,10],
			signifDigits(quantileY[1,10], quantileY[2,10]))
	cat(sprintf("99%% (%8g, %8g)\tk = %8g  \n",L,R,signif((R-L)/(2*sdY[1]),2)))
	
	L = signif(quantileY[1,2],
			signifDigits(quantileY[1,2], quantileY[2,2]))
	R = signif(quantileY[1,9],
			signifDigits(quantileY[1,9], quantileY[2,9]))
	cat(sprintf("95%% (%8g, %8g)\tk = %8g  \n",L,R,signif((R-L)/(2*sdY[1]),2)))
	
	L = signif(quantileY[1,3],
			signifDigits(quantileY[1,3], quantileY[2,3]))
	R = signif(quantileY[1,8],
			signifDigits(quantileY[1,8], quantileY[2,8]))
	cat(sprintf("90%% (%8g, %8g)\tk = %8g  \n",L,R,signif((R-L)/(2*sdY[1]),2)))
	
	L = signif(quantileY[1,4],
			signifDigits(quantileY[1,4], quantileY[2,4]))
	R = signif(quantileY[1,7],
			signifDigits(quantileY[1,7], quantileY[2,7]))
	cat(sprintf("68%% (%8g, %8g)\tk = %8g  \n",L,R,signif((R-L)/(2*sdY[1]),2)))
	
	L = signif(quantileY[1,5],
			signifDigits(quantileY[1,5], quantileY[2,5]))
	R = signif(quantileY[1,6],
			signifDigits(quantileY[1,6], quantileY[2,6]))
	cat(sprintf("50%% (%8g, %8g)\tk = %8g  \n",L,R,signif((R-L)/(2*sdY[1]),2)))
	
	cat("\n--------------------------------------------\n")
	cat("Gauss's Formula (GUM's Linear Approximation)", "\n\n")
	cat("        y  =", signif(sdGauss[1],
					signifDigits(meanY[1], meanY[2])), "\n")
	cat("      u(y) =", signif(sdGauss[2],
					signifDigits(sdY[1], sdY[2])), "\n\n")
	print(budget)
	cat("============================================\n")
	
	if(outputFile!="no-output")
	{
		
		
		write("===== RESULTS ==============================\n", file = outputFile2, append = FALSE)
		write("Monte Carlo Method\n", file = outputFile2, append=TRUE)
		write(paste("Summary statistics for sample of size",
						nbReal, "\n"), file = outputFile2, append = TRUE)
		write(paste("ave     =", signif(meanY[1],
								signifDigits(meanY[1], meanY[2]))), file = outputFile2, append = TRUE)
		write(paste("sd      =", signif(sdY[1], signifDigits(sdY[1], sdY[2]))), file = outputFile2, append = TRUE)
		write(paste("median  =",
						signif(medianY[1], signifDigits(medianY[1], medianY[2]))), file = outputFile2, append = TRUE)
		write(paste("mad     =", signif(madY[1],
								signifDigits(madY[1], madY[2])), "\n"), file = outputFile2, append = TRUE)
		
		write("Coverage intervals\n", file = outputFile2, append = TRUE)
		L = signif(quantileY[1,1],
				signifDigits(quantileY[1,1], quantileY[2,1]))
		R = signif(quantileY[1,10],
				signifDigits(quantileY[1,10], quantileY[2,10]))
		write(sprintf("99%% (%8g, %8g)\tk = %8g",L,R,signif((R-L)/(2*sdY[1]),2)),
				file = outputFile2, append = TRUE)
		
		L = signif(quantileY[1,2],
				signifDigits(quantileY[1,2], quantileY[2,2]))
		R = signif(quantileY[1,9],
				signifDigits(quantileY[1,9], quantileY[2,9]))
		write(sprintf("95%% (%8g, %8g)\tk = %8g",L,R,signif((R-L)/(2*sdY[1]),2)),
				file = outputFile2, append = TRUE)
		
		L = signif(quantileY[1,3],
				signifDigits(quantileY[1,3], quantileY[2,3]))
		R = signif(quantileY[1,8],
				signifDigits(quantileY[1,8], quantileY[2,8]))
		write(sprintf("90%% (%8g, %8g)\tk = %8g",L,R,signif((R-L)/(2*sdY[1]),2)),
				file = outputFile2, append = TRUE)
		
		L = signif(quantileY[1,4],
				signifDigits(quantileY[1,4], quantileY[2,4]))
		R = signif(quantileY[1,7],
				signifDigits(quantileY[1,7], quantileY[2,7]))
		write(sprintf("68%% (%8g, %8g)\tk = %8g",L,R,signif((R-L)/(2*sdY[1]),2)),
				file = outputFile2, append = TRUE)
		
		L = signif(quantileY[1,5],
				signifDigits(quantileY[1,5], quantileY[2,5]))
		R = signif(quantileY[1,6],
				signifDigits(quantileY[1,6], quantileY[2,6]))
		write(sprintf("50%% (%8g, %8g)\tk = %8g",L,R,signif((R-L)/(2*sdY[1]),2)), file=outputFile2,
				append=TRUE)
		
		write("\n--------------------------------------------\n",
				file = outputFile2, append = TRUE)
		write(paste("Gauss's Formula (GUM's Linear Approximation)",
						"\n"), file = outputFile2, append = TRUE)
		write(paste("      y    =",
						signif(sdGauss[1], signifDigits(meanY[1],
										meanY[2]))), file = outputFile2, append = TRUE)
		write(paste("      u(y) =",
						signif(sdGauss[2],
								signifDigits(sdY[1], sdY[2])), "\n"),
				file = outputFile2, append = TRUE)
		capture.output(budget, file=outputFile2, append=TRUE)
		write("============================================",
				file = outputFile2, append = TRUE)
	}
}
